package vt.CS5934.SwitchRoom.jobs;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.EnableAsync;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import vt.CS5934.SwitchRoom.models.*;
import vt.CS5934.SwitchRoom.repositories.*;
import vt.CS5934.SwitchRoom.utility.LookupTables;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Stream;

import static vt.CS5934.SwitchRoom.utility.UsefulTools.DATE_FORMAT;


@Component
@EnableAsync
public class OfferWishlistMatchingJob {

    private final Logger logger = LoggerFactory.getLogger(OfferWishlistMatchingJob.class);

    @Autowired
    MatchingJobInfoRepository matchingJobInfoRepository;
    @Autowired
    UserOfferRepository userOfferRepository;
    @Autowired
    OfferWaitingMatchRepository offerWaitingMatchRepository;
    @Autowired
    WishlistItemRepository wishlistItemRepository;
    @Autowired
    WishlistWaitingMatchRepository wishlistWaitingMatchRepository;

    @Autowired
    MatchedWishlistRecordRepository matchedWishlistRecordRepository;


    @Async
    @Scheduled(fixedDelayString = "${offer.wishlist.job.gap.seconds:60}000",
            initialDelayString = "${offer.wishlist.job.init.delay.seconds:60}000")
    public void jobStarter(){
        logger.info("Offer Wishlist Job start ...");
        HashMap<Long, MatchingJobInfoModel> lastPullTimeMap = new HashMap<>();
        lastPullTimeMap.put(LookupTables.OFFER_JOB_DB_ID, null);
        lastPullTimeMap.put(LookupTables.WISHLIST_JOB_DB_ID, null);
        lastPullTimeMap.put(LookupTables.MATCHING_JOB_DB_ID, null);
        try{
            //check the job status table, get the last pull time
            getLastPullTime(lastPullTimeMap);
            logger.info("getLastPullTime() finished");

            // Query and push the new offer records into its waiting list
            pushNewOfferRecordsIntoWaitingTable(lastPullTimeMap.get(LookupTables.OFFER_JOB_DB_ID));
            logger.info("pushNewOfferRecordsIntoWaitingTable() finished");

            // Query and push the new wishlist records into its waiting list
            pushNewWishlistItemRecordsIntoWaitingTable(lastPullTimeMap.get(LookupTables.WISHLIST_JOB_DB_ID));
            logger.info("pushNewWishlistItemRecordsIntoWaitingTable() finished");

        }catch (Exception e){
            logger.error("Error occur when fetch the new records from Offer table / Wishlist table or" +
                    "push them into their waiting table: " +e.fillInStackTrace());
            // may need to stop the task if needed.
        }

        try{
            //find the match records and push to the matched table
            findMatchFromWaitingTable(lastPullTimeMap.get(LookupTables.MATCHING_JOB_DB_ID));
        }catch (Exception e){
            logger.error("Error occur when finding matches and push them to matched table: " +e.fillInStackTrace());
        }

    }




    /**
     * This function is to get the last pull time into a map, so the job knows the starting time range.
     * If DB do not have the job record info, it will build a default with starting of time as last pull time
     * @param lastPullTimeMap a map with job ids
     */
    private void getLastPullTime(HashMap<Long, MatchingJobInfoModel> lastPullTimeMap){
        logger.info("Starting getLastPullTime function");
        for(Long jobId : lastPullTimeMap.keySet()){
            MatchingJobInfoModel jobInfo = matchingJobInfoRepository.findByJobId(jobId);
            logger.info("Got job with id: {} info, the record is: {}", jobId, jobInfo);
            if(jobInfo == null){
                logger.info("Offer job info is null, building initial offer job info record");
                jobInfo = new MatchingJobInfoModel(jobId,LookupTables.JOB_STATUS.Running, new Date(0));
            }
            logger.info("The last pull time for job with id: {} is: {}",
                    jobId, DATE_FORMAT.format(jobInfo.getLastPullTime()));
            jobInfo.setJobStatus(LookupTables.JOB_STATUS.Running);
            jobInfo = matchingJobInfoRepository.save(jobInfo);
            lastPullTimeMap.put(jobId,jobInfo);
        }
    }

    /**
     * This function will pull the new Offer records from Offer_table and push them into offer_waiting_table
     * @param offerJobRecord offer job information to start the task
     */
    private void pushNewOfferRecordsIntoWaitingTable(MatchingJobInfoModel offerJobRecord){
        logger.info("Starting pushNewOfferRecordsIntoWaitingTable function with lastPullTime:{}"
                ,DATE_FORMAT.format(offerJobRecord.getLastPullTime()));

        // update the job info record time
        Date lastPullOfferTime = offerJobRecord.getLastPullTime();
        offerJobRecord.setJobStatus(LookupTables.JOB_STATUS.Done);
        offerJobRecord.setLastPullTime(new Date(System.currentTimeMillis()));

        // fetch the Offer records and push them to waiting table
        List<UserOfferModel> newOfferList = userOfferRepository
                .findAllByOfferingAndModifyDateAfter(true, lastPullOfferTime);
        logger.info("Fetched {} new Offer records, and set the last pull time to {}", newOfferList.size(),
                DATE_FORMAT.format(offerJobRecord.getLastPullTime()));
        for(UserOfferModel newOffer : newOfferList){
            OfferWaitingMatchModel offerWaitingRecord = new OfferWaitingMatchModel(newOffer.getUserId(),
                    newOffer.getAvailableTimeStart(), newOffer.getAvailableTimeEnd(), newOffer.getStateCityCode());
            offerWaitingMatchRepository.save(offerWaitingRecord);
        }

        matchingJobInfoRepository.save(offerJobRecord);
    }

    /**
     * This function will pull the new wishlist records from wishlist_table and push them into wishlist_waiting_table
     * @param wishlistJobInfo wishlist job information to start the task
     */
    private void pushNewWishlistItemRecordsIntoWaitingTable(MatchingJobInfoModel wishlistJobInfo){
        logger.info("Starting pushNewWishlistItemRecordsIntoWaitingTable function with lastPullTime:{}"
                ,DATE_FORMAT.format(wishlistJobInfo.getLastPullTime()));

        // update the job info record time
        Date lastPullOfferTime = wishlistJobInfo.getLastPullTime();
        wishlistJobInfo.setJobStatus(LookupTables.JOB_STATUS.Done);
        wishlistJobInfo.setLastPullTime(new Date(System.currentTimeMillis()));

        // fetch the Wishlist records and push them to waiting table
        List<WishlistItemModel> newWishlistItemList = wishlistItemRepository
                .findAllByModifyDateAfter(lastPullOfferTime);
        logger.info("Fetched {} new Offer records, and set the last pull time to {}", newWishlistItemList.size(),
                DATE_FORMAT.format(wishlistJobInfo.getLastPullTime()));
        for(WishlistItemModel newWishlistItem : newWishlistItemList){
            WishlistWaitingMatchModel wishlistWaitingRecord = new WishlistWaitingMatchModel(
                    newWishlistItem.getWishlistItemId(), newWishlistItem.getStartTime()
                    , newWishlistItem.getEndTime(), newWishlistItem.getStateCityCode());
            wishlistWaitingMatchRepository.save(wishlistWaitingRecord);
        }

        matchingJobInfoRepository.save(wishlistJobInfo);
    }

    /**
     * This function will check records in the waiting table and try to match them. if match find, it will push them
     * into matched table.
     * @param matchJobRecord job info
     */
    private void findMatchFromWaitingTable(MatchingJobInfoModel matchJobRecord){
        logger.info("Starting findMatchFromWaitingTable function with lastPullTime:{}"
                ,DATE_FORMAT.format(matchJobRecord.getLastPullTime()));
        // update the job info record time
        Date lastPullOfferTime = matchJobRecord.getLastPullTime();
        matchJobRecord.setJobStatus(LookupTables.JOB_STATUS.Done);
        matchJobRecord.setLastPullTime(new Date(System.currentTimeMillis()));

        // fetch both new records
        List<OfferWaitingMatchModel> newOfferRecords =
                offerWaitingMatchRepository.findAllByModifyDateAfter(lastPullOfferTime);
        List<WishlistWaitingMatchModel> newWishlistRecords =
                wishlistWaitingMatchRepository.findAllByModifyDateAfter(lastPullOfferTime);

        // if one of them have new records
        if(newOfferRecords.size() > 0 || newWishlistRecords.size() > 0){
            // load new Offers into map
            HashMap<Long, List<OfferWaitingMatchModel>> newStateCityCodeOfferMap = new HashMap<>();
            HashMap<Long, List<WishlistWaitingMatchModel>> newStateCityCodeWishlistMap = new HashMap<>();
            loadListIntoMap(newOfferRecords, newWishlistRecords,newStateCityCodeOfferMap, newStateCityCodeWishlistMap);

            // fetch both old records
            List<OfferWaitingMatchModel> oldOfferRecords =
                    offerWaitingMatchRepository.findAllByModifyDateLessThanEqualAndStateCityCodeIn(lastPullOfferTime,
                            newStateCityCodeWishlistMap.keySet());
            List<WishlistWaitingMatchModel> oldWishlistRecords =
                    wishlistWaitingMatchRepository.findAllByModifyDateLessThanEqualAndStateCityCodeIn(lastPullOfferTime,
                            newStateCityCodeOfferMap.keySet());

            // build maps by use the StateCityCode as the key so make the next pair step easier.
            HashMap<Long, List<OfferWaitingMatchModel>> oldStateCityCodeOfferMap = new HashMap<>();
            HashMap<Long, List<WishlistWaitingMatchModel>> oldStateCityCodeWishlistMap = new HashMap<>();
            loadListIntoMap(oldOfferRecords, oldWishlistRecords,oldStateCityCodeOfferMap, oldStateCityCodeWishlistMap);

            // after fetch all new Wishlist item records by unique stateCityCode
            // match them with all Offer records
            newStateCityCodeWishlistMap.entrySet()
                    .parallelStream()
                    .forEach(entry -> {
                        List<OfferWaitingMatchModel> newStateCityCodeOfferList =
                                !newStateCityCodeOfferMap.containsKey(entry.getKey())? new ArrayList<>()
                                : newStateCityCodeOfferMap.get(entry.getKey());
                        List<OfferWaitingMatchModel> oldStateCityCodeOfferList =
                                !oldStateCityCodeOfferMap.containsKey(entry.getKey())? new ArrayList<>()
                                :oldStateCityCodeOfferMap.get(entry.getKey());
                        List<OfferWaitingMatchModel> offerList =
                                Stream.concat(newStateCityCodeOfferList.stream(),
                                        oldStateCityCodeOfferList.stream()).toList();
                        findMatchedOfferForNewWishlistItemToDB(offerList, entry.getValue());
                    });


            // If new offers and old Wishlist item records' unique stateCityCode has intersection,
            // fetch old Wishlist item by new Offer stateCityCode
            if(newOfferRecords.size() > 0){
                oldStateCityCodeWishlistMap.entrySet()
                        .parallelStream()
                        .forEach(entry ->{
                            List<OfferWaitingMatchModel> newStateCityCodeOfferList =
                                    !newStateCityCodeOfferMap.containsKey(entry.getKey())? new ArrayList<>()
                                            : newStateCityCodeOfferMap.get(entry.getKey());
                            findMatchedOfferForNewWishlistItemToDB(newStateCityCodeOfferList,
                                    entry.getValue());
                        });
            }

        }

        matchingJobInfoRepository.save(matchJobRecord);
    }


    private void findMatchedOfferForNewWishlistItemToDB(List<OfferWaitingMatchModel> offerList,
                                  List<WishlistWaitingMatchModel> wishlistItemList) {
        wishlistItemList.stream().forEach( wishlistItem ->{
            useWishlistItemAndOfferListBuildMatchedRecordDB(wishlistItem, offerList);
        });

    }

    private void useWishlistItemAndOfferListBuildMatchedRecordDB(WishlistWaitingMatchModel wishlistItem,
                                                               List<OfferWaitingMatchModel> offerList) {
        List<MatchedWishlistRecordModel> matchedRecordList = new ArrayList<>();
        offerList.stream().forEach( offer ->{
            if(isOverlapping(wishlistItem.getStartTime(),wishlistItem.getEndTime(),
                    offer.getStartTime(),offer.getEndTime()))
                matchedRecordList.add(new MatchedWishlistRecordModel(wishlistItem.getWishlistItemId(),
                        offer.getOfferId(), LookupTables.MATCHING_RECORD_USER_RESULT.Waiting,
                        LookupTables.MATCHING_RECORD_USER_RESULT.Waiting));
        });
        matchedWishlistRecordRepository.saveAll(matchedRecordList);
    }

    private void loadListIntoMap(List<OfferWaitingMatchModel> offerRecords,
                                 List<WishlistWaitingMatchModel> wishlistRecords,
                                 HashMap<Long, List<OfferWaitingMatchModel>> stateCityCodeOfferMap,
                                 HashMap<Long, List<WishlistWaitingMatchModel>> stateCityCodeWishlistMap){
        offerRecords.stream().forEach(a -> {
            if(!stateCityCodeOfferMap.containsKey(a.getStateCityCode()))
                stateCityCodeOfferMap.put(a.getStateCityCode(), new ArrayList<>());
            stateCityCodeOfferMap.get(a.getStateCityCode()).add(a);
        });

        wishlistRecords.stream().forEach(a->{
            if(!stateCityCodeWishlistMap.containsKey(a.getStateCityCode()))
                stateCityCodeWishlistMap.put(a.getStateCityCode(), new ArrayList<>());
            stateCityCodeWishlistMap.get(a.getStateCityCode()).add(a);
        });
    }

    private boolean isOverlapping(Date start1, Date end1, Date start2, Date end2) {
        return start1.before(end2) && start2.before(end1);
    }


}
